/* -*- Mode: C -*-
 *Header:
 *File: object.C (for C++ & cc)
 *Author: Noda Itsuki
 *Date: 1995/02/21
 *EndHeader:
 */

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
//#include <windef.h>

#include "param.h"
#include "types.h"
#include "utility.h"
#include "object.h"
#include "field.h"


/*
 *===================================================================
 *Part: Plane Vector
 *===================================================================
 */

PVector::PVector(const double vx, const double vy)
{
	x = vx ;
	y = vy ;
}

void PVector::operator +=(const PVector& v)
{
	x += v.x ;
	y += v.y ;
}

void PVector::operator -=(const PVector& v)
{
	x -= v.x ;
	y -= v.y ;
}

void PVector::operator *=(const double& a)
{
	x *= a ;
	y *= a ;
}

void PVector::operator /=(const double& a)
{
	x /= a ;
	y /= a ;
}

ostream& operator<< (ostream& o, const PVector& v) 
{
	return o << "#V[" << v.x << "," << v.y << "]" ;
}

PVector operator +(const PVector& a, const PVector& b)
{
	return PVector((a.x + b.x), (a.y + b.y)) ;
}

PVector operator -(const PVector& a, const PVector& b)
{
	return PVector((a.x - b.x), (a.y - b.y)) ;
}

void PVector::normalize(const Value& l) 
{
	*this *= (l/mmax(r(),EPS)) ;
}

Value PVector::distance(const PVector& orig)
{
	return (*this - orig).r() ;
}

Angle PVector::angle()
{
	return Atan(y,x) ;
}

Angle PVector::angle(const PVector& dir)
{
	Angle ang = Atan(dir.y,dir.x) - Atan(y,x) ;

	return normalize_angle(ang) ;
}

void PVector::rotate(const Angle& ang)
{
	Value r1 = r() ;
	Angle th1 = th() ;

	x = r1 * cos(th1 + ang) ;
	y = r1 * sin(th1 + ang) ;
}

Angle PVector::vangle(const PVector& target, const PVector& origin) 
{
	PVector axis = origin - *this ;
	PVector dir = target - *this ;
	return axis.angle(dir) ;
}

Angle PVector::vangle(const PVector& target, const Angle& origin)
{
	PVector dir = target - *this ;
	Angle ang = dir.angle() - origin ;

	return normalize_angle(ang) ;
}

/*
 *===================================================================
 *Part: Area
 *===================================================================
 */

RArea::RArea(const Value l, const Value r, const Value t, const Value b) 
{
	left = l ; right = r ;
	top = t ; bottom = b ;
}

RArea::RArea(const PVector center,const PVector size)
{
	left = center.x - size.x/2.0 ;
	right = center.x + size.x/2.0 ;
	top = center.y - size.y/2.0 ;
	bottom = center.y + size.y/2.0 ;
}

Logical RArea::inArea(const PVector &p)
{
	return (p.x >= left) && (p.x <= right) && (p.y >= top) && (p.y <= bottom) ;
}

PVector RArea::nearestHEdge(const PVector &p) 
		/* find nearest horizontal line */
{
	static PVector r ;

	r.x = mmin(mmax(p.x,left),right) ;
	r.y = (Abs(p.y - top) < Abs(p.y - bottom)) ? top : bottom ;

	return r ;
}

PVector RArea::nearestVEdge(const PVector &p) 
		/* find nearest vertical line */
{
	static PVector r ;

	r.x = (Abs(p.x - left) < Abs(p.x - right)) ? left : right ;
	r.y = mmin(mmax(p.y,top),bottom) ;

	return r ;
}

PVector RArea::nearestEdge(const PVector &p) 
{
	if(mmin(Abs(p.x-left),Abs(p.x-right)) < mmin(Abs(p.y-top),Abs(p.y-bottom)))
		return nearestVEdge(p) ;
	else
		return nearestHEdge(p) ;
}

PVector RArea::randomize() 
{
	static PVector r ;

	r.x = drand(left, right) ;
	r.y = drand(bottom, top) ;

	return r ;
}

ostream& operator<< (ostream& o, const RArea& a) 
{
	return o << "#A[h:" << a.left << "~" << a.right 
			<< ",v:" << a.top << "~" << a.bottom << "]" ;
}

CArea::CArea(const PVector cnt, const Value radius)
{
	center = cnt ;
	r = radius ;
}

Logical CArea::inArea(const PVector &p)
{
	return r >= center.distance(p) ;
}

PVector CArea::nearestEdge(const PVector &p)
{
	PVector dif = p - center ;
	if (dif.x == 0.0 && dif.y == 0.0)
		dif = PVector(EPS, EPS) ;

	dif.normalize(r) ;

	return center + dif ;
}

ostream& operator<< (ostream& o, const CArea& a) 
{
	return o << "#A[x:" << a.center.x <<
				",y:" << a.center.y <<
				",r:" << a.r << "]" ;
}

/*
 *===================================================================
 *Part: Visual Info
 *===================================================================
 */

VisualInfo::VisualInfo(double version_tmp)
{
	object = (PObject *)NULL ;
	version = version_tmp ;
}

VisualInfo::VisualInfo()
{
}

void VisualInfo::set(double version_tmp)
{
	object = (PObject *)NULL ;
	version = version_tmp ;
}


ostream& operator<< (ostream& o, const VisualInfo& vi)
{
	Value ed ;
	Value rdist_change ;
	Angle rdir_change ;
	char tmp[128] ;
				/* modified distance */
	ed = Quantize(exp(Quantize(log(vi.distance + EPS),vi.qstep)),0.1) ;

	sscanf(vi.name,"(%s",tmp) ;
	if (!strcmp(tmp,"player")) {
		double prob ;
		prob = (ed-TEAM_FAR_LENGTH)/(TEAM_TOOFAR_LENGTH-TEAM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
			o << "(" << ((Player *)(vi.object))->name_toofar << " " ;
			if (vi.highquality)
				o << ed << " " ; 
			return o << Rad2IDeg(vi.direction) << ")" ;
		}

		prob = (ed-UNUM_FAR_LENGTH)/(UNUM_TOOFAR_LENGTH-UNUM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
			o << "(" << ((Player *)(vi.object))->name_far << " " ;
			if (vi.highquality)
				o << ed << " " ;
			return o << Rad2IDeg(vi.direction) << ")" ;
		}
		o << "(" << vi.name << " " ;

		if (vi.highquality) {
			o << ed << " " ;
			if (vi.distance != Zero)
				rdist_change = ed * Quantize(vi.dist_change/vi.distance, 0.02) ;
			else
				rdist_change = Zero ;
			if (vi.dir_change != Zero)
				rdir_change = Quantize(vi.dir_change, 0.1) ;
			else
				rdir_change = Zero ;
			if (vi.version >= 5.0)
				return o << Rad2IDeg(vi.direction) << " "
						<< rdist_change << " "
						<< rdir_change << " "
						<< Rad2IDeg(vi.body_dir) << " "
						<< Rad2IDeg(vi.face_dir) << ")" ;
			if (vi.version > 4.0)
				return o << Rad2IDeg(vi.direction)  << " "
						<< rdist_change << " " 
						<< rdir_change << " "
						<< Rad2IDeg(vi.body_dir) << ")" ;
			if (vi.version > 3.0)
				return o << Rad2IDeg(vi.direction) << " "
						<< rdist_change << " " 
						<< rdir_change << " "
						<< Rad2IDeg(vi.body_dir) << ")" ;
			else
				return o << Rad2IDeg(vi.direction) << " "
						<< rdist_change << " " 
						<< rdir_change << ")" ;
		}
		else
			return o << Rad2IDeg(vi.direction) << ")" ;
	}
	else if (!strcmp(tmp, "line")) {
		o << "(" << vi.name << " " ;
		if (vi.highquality)
			o << ed << " " ;
		return o << Rad2IDeg(vi.direction) << ")" ;
	}
	else if (!strcmp(tmp, "ball)") || !strcmp(tmp, "flag")) {
		double prob ;
		prob = (ed-UNUM_FAR_LENGTH)/(UNUM_TOOFAR_LENGTH-UNUM_FAR_LENGTH) ;
		if (prob>drand(0,1)) {
			o << "(" << vi.name << " " ;
			if (vi.highquality)
				o << ed << " " ;
			return o << Rad2IDeg(vi.direction) << ")" ;
		}
	}
	else {		/* near object that can't see */
		o << "(" << vi.name << " " ;
		if (vi.highquality)
			o << ed << " " ;
		return o << Rad2IDeg(vi.direction) << ")" ;
	}
	o << "(" << vi.name << " " ;
	if (vi.highquality) {
		o << ed << " " ;
		if (vi.distance != Zero)
			rdist_change = ed * Quantize(vi.dist_change/vi.distance, 0.02) ;
		else
			rdist_change = Zero ;
		if (vi.dir_change != Zero)
			rdir_change = Quantize(vi.dir_change, 0.1) ;
		else
			rdir_change = Zero ;
		return o << Rad2IDeg(vi.direction) << " "
				<< rdist_change << " " << rdir_change << ")" ;
	}
	else
		return o << Rad2IDeg(vi.direction) << ")" ;
}

/*
 *===================================================================
 *Part: PObject 
 *===================================================================
 */

TheNumber PObject::Max = 0 ;

PObject::PObject(const Name nm,const PVector& p, const Value v) 
{
	id = Max ;
	Max++ ;
	name = nm ;
	stadium = (Stadium *)NULL ;
	vi = (VisualInfo *)NULL ;
	pos = p ;
	obj_ver = v ;
	size = 0.0 ;
	angle = 0.0 ;
	enable = TRUE ;
}

ostream& operator<< (ostream& o, const PObject& obj) 
{
	o << "#Ob[" << obj.id ;
	if( obj.name != NULLCHAR )
		o << ":" << obj.name << "" ;
	return o 
		<< ":pos=" << obj.pos 
		<< ",size=" << obj.size 
		<< ",angle=" << obj.angle
		<< "]" ;
}

void PObject::Set(const PVector& p, const Value& s)
{
	pos = p ;
	size = s ;
}

Angle PObject::vangle(const PObject& obj)
{
	return pos.vangle(obj.pos,angle) ;
}

int Player::scanField()
{
	PObjectTable *tbl ;
	MPObjectTable *mtbl ;
	int i,j,c ;
	Angle ang, angle_sight ;
	char tmp[16] ;
	PVector etmp, vtmp ;

	if(!vi) 
		vi = new VisualInfo[MaxObject];
		for(int z=0;z<MaxObject;z++) vi[z].set(version);
	c = 0 ;

	tbl = &(stadium->votable) ;
	mtbl = &(stadium->motable) ;

	angle_sight = angle + angle_neck_committed ;

	for(i = 0 ; i < tbl->n ; i++ ) {
		if(tbl->object[i] == this || (*tbl->object[i]).enable == FALSE)
			continue ;
		ang = this->vangle(*(tbl->object[i])) - angle_neck_committed ;
		if(fabs(ang) < vis_angle/2.0) {
			vi[c].highquality = highquality ;
			vi[c].object = tbl->object[i] ;
			vi[c].name = tbl->object[i]->name ;
			vi[c].distance = this->pos.distance(tbl->object[i]->pos) ;
			vi[c].direction = ang ;

			if (vi[c].distance == Zero)
				vi[c].dist_change = vi[c].dir_change = Zero ;
			else {
				etmp = tbl->object[i]->pos - this->pos ;
				sscanf(tbl->object[i]->name, "(%[^ )]", tmp);
				if (!strcmp(tmp, "ball")) {
					vi[c].qstep = stadium->dist_qstep ;
					for(j = 0 ; j < mtbl->n ; j++ ) {
						if (tbl->object[i]->id == mtbl->object[j]->id) {
							vtmp = mtbl->object[j]->vel - this->vel ;
							break ;
						}
					}
				}
				else if (!strcmp(tmp, "player")) {
					vi[c].qstep = stadium->dist_qstep ;
					for(j = 0 ; j < mtbl->n ; j++ ) {
						if (tbl->object[i]->id == mtbl->object[j]->id) {
							vtmp = mtbl->object[j]->vel - this->vel ;
							break ;
						}
					}
					vi[c].body_dir = 
						normalize_angle(tbl->object[i]->angle - (this->angle + this->angle_neck_committed)) ;
					vi[c].face_dir = 
						normalize_angle(
							(((Player *)tbl->object[i])->angle_neck_committed
								+ ((Player *)tbl->object[i])->angle)
								- (this->angle + this->angle_neck_committed)) ;

				}
				else {
					vi[c].qstep = stadium->land_qstep ;
					vtmp.x = -this->vel.x ;
					vtmp.y = -this->vel.y ;
				}
				etmp /= vi[c].distance ;
				vi[c].dist_change = vtmp.x * etmp.x + vtmp.y * etmp.y ;
				vi[c].dir_change = RAD2DEG * (vtmp.y * etmp.x
								- vtmp.x * etmp.y) / vi[c].distance ;
			}
			c++ ;
		}
		else if (this->pos.distance(tbl->object[i]->pos) <= vis_distance){
			vi[c].highquality = highquality ;
			vi[c].object = tbl->object[i] ;
			sscanf(tbl->object[i]->name, "(%s", tmp);
			if (!strcmp(tmp, "flag")) {
				vi[c].name = O_TYPE_FLAG_NAME ;
				vi[c].qstep = stadium->land_qstep ;
			}
			else if (!strcmp(tmp, "goal")) {
				vi[c].name = O_TYPE_GOAL_NAME ;
				vi[c].qstep = stadium->land_qstep ;
			}
			else if (!strcmp(tmp, "ball)")) {
				vi[c].name = O_TYPE_BALL_NAME ;
				vi[c].qstep = stadium->dist_qstep ;
			}
			else if (!strcmp(tmp, "player")) {
				vi[c].name = O_TYPE_PLAYER_NAME ;
				vi[c].qstep = stadium->dist_qstep ;
			}
			else
				vi[c].name = "unknown" ;

			vi[c].distance = this->pos.distance(tbl->object[i]->pos) ;
			vi[c].direction = ang ;
			c++ ;
		}
	}

	 // line information
	ang = Rad2IDeg(normalize_angle(angle_sight)) ;
	PObject *dir[2], *dir2[2] ;
	Value edge[2], r[4] ;
	int out_of_field = TRUE ;
	int even = FALSE ;

#ifndef WRONG_LINE_VIEW		/* line view type is normal */
	if (pos.y < -PITCH_WIDTH/2) {	/* above line_t */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
			}
			else if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
			}
			else if (ang == 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_t) ;
				if (ang < 90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_t) ;
				dir2[i] = &(stadium->field.line_b) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;
		}
	}
	else if (pos.y > PITCH_WIDTH/2) {	/* under line_b */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > -180 && ang < -90) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_r
			}
			else if (ang > -180 && ang < -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;// to line_l
			}
			else if (ang == -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_b
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_b) ;
				if (ang > -90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_b) ;
				dir2[i] = &(stadium->field.line_t) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;

		}
	}
	else if (pos.x < -PITCH_LENGTH/2) {	/* left of line_l */
		if (ang > -90 && ang < 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_t
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else if (ang > 0 && ang < 90) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;   // to line_b
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else if (ang == 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_l) ;
			if (ang < 0)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_l) ;
			dir2[i] = &(stadium->field.line_r) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else if (pos.x > PITCH_LENGTH/2) {	/* right of line_r */
		if (ang > -180 && ang < -90) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;  // to line_t
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else if (ang > 90 && ang < 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle_sight)) ;   // to line_b
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else if (ang == 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ;  // to line_r
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle_sight)) ; // to line_l
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_r) ;
			if (ang < -90)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_r) ;
			dir2[i] = &(stadium->field.line_l) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else {		/* In the field */
		out_of_field = FALSE ;
		if (ang > -90 && ang <= 90) {
			dir[0] = &(stadium->field.line_r) ;
			edge[0] = PITCH_LENGTH/2 ;
		}
		else {
			dir[0] = &(stadium->field.line_l) ;
			edge[0] = -PITCH_LENGTH/2 ;
		}
		if (ang > -180 && ang <= 0) {
			dir[1] = &(stadium->field.line_t) ;
			edge[1] = -PITCH_WIDTH/2 ;
		}
		else {
			dir[1] = &(stadium->field.line_b) ;
			edge[1] = PITCH_WIDTH/2 ;
		}

		r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle_sight)) ;
		r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle_sight)) ;

		i = (r[0] <= r[1]) ? 0 : 1 ;
	}
#else		/* defined WRONG_LINE_VIEW */
	out_of_field = FALSE ;
	if (ang > -90 && ang <= 90) {
		dir[0] = &(stadium->field.line_r) ;
		edge[0] = PITCH_LENGTH/2 ;
	}
	else {
		dir[0] = &(stadium->field.line_l) ;
		edge[0] = -PITCH_LENGTH/2 ;
	}
	if (ang > -180 && ang <= 0) {
		dir[1] = &(stadium->field.line_t) ;
		edge[1] = -PITCH_WIDTH/2 ;
	}
	else {
		dir[1] = &(stadium->field.line_b) ;
		edge[1] = PITCH_WIDTH/2 ;
	}

	r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle_sight)) ;
	r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle_sight)) ;
	i = (r[0] <= r[1]) ? 0 : 1 ;
#endif

	if (i == 0)
		ang = -angle_sight + ((angle_sight > 0) ? PI/2 : -PI/2) ;
	else
		ang = -angle_sight ;
	if (ang > PI/2) ang -= PI ;
	if (ang < -PI/2) ang += PI ;

	vi[c].highquality = highquality ;
	vi[c].object = dir[i] ;
	vi[c].name = dir[i]->name ;
	vi[c].distance = r[i] ;
	vi[c].direction = ang ;
	vi[c].qstep = stadium->land_qstep ;
	c++ ;

	if (out_of_field) {
		vi[c].highquality = highquality ;
		vi[c].object = dir2[i] ;
		vi[c].name = dir2[i]->name ;
		vi[c].distance = r[3-i] ;
		vi[c].qstep = stadium->land_qstep ;
		if (even)
			vi[c].direction = ang ;
		else
			vi[c].direction = ang + ((ang > 0) ? -PI/2 : PI/2) ;
		c++ ;
	}

	return c ;

}


/*
 *===================================================================
 *Part: MPObject
 *===================================================================
 */

MPObject::MPObject(const Name nm) : PObject(nm) { }

ostream& operator<< (ostream& o, const MPObject& obj) 
{
	o << "#Ob[" << obj.id ;
	if( obj.name != NULLCHAR )
		o << ":" << obj.name << "" ;
	return o 
		<< ":pos=" << obj.pos 
	 	<< ",size=" << obj.size
	 	<< ",angle=" << Rad2IDeg(obj.angle)
		<< ",vel=" << obj.vel 
		<< ",acc=" << obj.acc 
		<< ",decay=" << obj.decay
		<< ",randp=" << obj.randp
		<< "]" ;
}

void MPObject::Set(
				const MPObjectType& t,
				const PVector& p,
				const Value& s,
				const Angle& ang, 
				const PVector& v,
				const PVector& a,
				const Value& d,
				const Value& r)
{
	type = t ;
	pos = p ;
	size = s ;
	angle = ang ;
	vel = v ;
	acc = a ;
	decay = d ;
	randp = r ;
}

PVector MPObject::noise() 
{
	Value maxrnd = randp * vel.r() ;
	return PVector(drand(-maxrnd,maxrnd),drand(-maxrnd,maxrnd)) ;
}

PVector MPObject::wind()
{
	return PVector(vel.r() * (weather->wind_vector.x +
			drand(-weather->wind_rand, weather->wind_rand) ) /
			(weight * WIND_WEIGHT), vel.r() * (weather->wind_vector.y +
			drand(-weather->wind_rand, weather->wind_rand)) /
			(weight * WIND_WEIGHT)) ;
}

void MPObject::_inc() 
{
	Value tmp ;

	if (acc.x || acc.y) {
		vel += acc ;
		if ((tmp = vel.r()) > max_speed)
			vel *= (max_speed / tmp) ;
	}
	if (this->type == MPO_Player) {
                angle = ((Player *)this)->angle_body ;
                ((Player *)this)->angle_neck_committed = ((Player *)this)->angle_neck ;
	}
	vel += noise() ;
	vel += wind() ;
	pos += vel ;
	vel *= decay ;
	acc *= Zero ;
}

void MPObject::_turn()
{
    if (this->type == MPO_Player) {
                angle = ((Player *)this)->angle_body ;
                ((Player *)this)->angle_neck_committed = ((Player *)this)->angle_neck ;
	}
}

Logical MPObject::collisionP(const MPObject& obj) 
{
	if(&obj == this) return FALSE ;
	if(&obj.enable == FALSE) return FALSE  ;
	return this->pos.distance(obj.pos) < (size + obj.size) ;
}

void MPObject::collisionCheck(MPObjectTable& tbl)
{
	tbl.collisionCheck(this) ;
}

void MPObjectTable::collisionCheck(MPObject* obj)
{
	TheNumber i ;
	Logical colp = FALSE ;
	MPObject *ob ;

	for(i = 0 ; i < n ; i++) {
		ob = object[i] ;
		if(obj->collisionP(*ob)) {
			colp = TRUE ;
			obj->collide(*ob) ;
			if (obj->type == MPO_Ball && ob->type == MPO_Player) {
				Player *player = (Player *)ob ;
				player->alive |= BALL_TO_PLAYER ;
				player->stadium->last_touch = player ;
			}
			else if (obj->type == MPO_Player && ob->type == MPO_Ball) {
				Player *player = (Player *)obj ;
				player->alive |= PLAYER_TO_BALL ;
				player->stadium->last_touch = player ;
			}
		}
	}
	if(colp)
		obj->vel *= -0.1 ;
}

void MPObject::collide(MPObject& obj)
{
	Value r = size + obj.size ;
	PVector dif = (pos - obj.pos) ;
	Value d = pos.distance(obj.pos) ;
	Angle th = fabs(dif.angle(vel)) ;
	Value l1 = d * cos(th) ;
	Value h = d * sin(th) ;
	Value cosp = h / r ;
	Value sinp = sqrt(1.0 - square(cosp)) ;
	Value l2 = r * sinp ;
	PVector dv = vel ;

	dv.normalize(-(l1 + l2)) ;

	pos += dv ;
}

void MPObject::step(MPObjectTable& tbl)
{
	++(*this) ;
	this->collisionCheck(tbl) ;
}

void MPObject::push(PVector f)
{
	acc = f ;
}

void MPObject::rotate(Angle a)
{
	angle += a ;
}

void MPObject::rotate_to(Angle dir)
{
	angle = dir ;
}

void MPObject::dash(Value f)
{
	push(Polar2PVector(f, angle)) ;
}

/*
 *===================================================================
 *Part: PObjectTable
 *===================================================================
 */

PObjectTable::PObjectTable(const TheNumber max)
{
	Max = max ;
	object = new PObject*[Max] ;
	n = 0 ;
}

PObjectTable::~PObjectTable()
{
	delete object ;
}

ostream& operator<< (ostream& o, const PObjectTable& tbl) 
{
	TheNumber i ;

	o << "#PObjectTable[n=" << tbl.n << "/" << tbl.Max << endl ;

	for(i = 0 ; i < tbl.n ; i++)
		o << "  " << *(tbl.object[i]) << endl ;

	return o << "]" << endl ;
}

PObject* PObjectTable::AssignObject(PObject* obj)
{
	object[n] = obj ;
	n++ ;
	return obj ;
}

PObject* PObjectTable::NewObject(const Name nm)
{
	PObject *obj = new PObject(nm) ;

	AssignObject(obj) ;

	return obj ;
}

/*
 *===================================================================
 *Part: MPObjectTable
 *===================================================================
 */
MPObjectTable::MPObjectTable(const TheNumber max)
{
	randomize() ;
	Max = max ;
	object = new MPObject*[Max] ;
	n = 0 ;
}

MPObjectTable::~MPObjectTable()
{
	delete object ;
}

ostream& operator<< (ostream& o, const MPObjectTable& tbl) 
{
	TheNumber i ;

	o << "#MPObjectTable[n=" << tbl.n << "/" << tbl.Max << endl ;

	 for(i = 0 ; i < tbl.n ; i++)
		o << "\t" << *(tbl.object[i]) << endl ;

	return o << "]" << endl ;
}

MPObject* MPObjectTable::AssignObject(MPObject* obj)
{
	object[n] = obj ;
	n++ ;
	return obj ;
}

MPObject* MPObjectTable::NewObject(const Name nm)
{
	MPObject *obj = new MPObject(nm) ;

	AssignObject(obj) ;

	return obj ;
}

void MPObjectTable::_inc()
{
	TheNumber i ;
	int sw ;
	MPObject **obj ;
	MPObject *buf ;

	obj = object ;

	/* randomize order */
	for(i = 0; i < n ; i++) {
		sw = irand(n) ;
		buf = *(obj+i) ;
		*(obj+i) = *(obj+sw) ;
		*(obj+sw) = buf ;
	}

	for(i = 0; i < n ; i++) {
		(**obj).step(*this) ;
		obj++ ;
	}
}

void MPObjectTable::_turn()
{
	TheNumber i ;
	int sw ;
	MPObject **obj ;
	MPObject *buf ;

	obj = object ;

	/* randomize order */
	for(i = 0; i < n ; i++) {
		sw = irand(n) ;
		buf = *(obj+i) ;
		*(obj+i) = *(obj+sw) ;
		*(obj+sw) = buf ;
	}

	for(i = 0; i < n ; i++) {
		(**obj)._turn() ;
		obj++ ;
	}
}

/*
 *===================================================================
 *Part: Player & Team
 *===================================================================
 */

Player::Player(Team *tm, Stadium *std, ID number) : MPObject() 
{
	stadium = std ;
	team = tm ;
	unum = number ;
	name = new char[128];
	name_far = new char[128];
	name_toofar = new char[128];
	stamina = std->stamina_max ;
	inertia_moment = std->inertia_moment ;
	recovery = 1.0 ;
	effort = 1.0 ;
	vis_angle = defangle = std->visangle ;
	vis_distance = std->visdist ;
	vis_send = 4 ;
	sendcnt = 0 ;
	highquality = TRUE ;
	weight = std->pweight ;
	weather = &(std->weather) ;
	max_speed = std->pspeed_max ;
	command_done = FALSE ;
	turn_neck_done = FALSE ;
	offside_mark = FALSE ;
	goalie = FALSE ;
	alive = DISABLE ;
	goalie_catch_ban = FALSE ;
	goalie_catch_probability = std->catch_prob ;
	hear_capacity_from_teammate = std->hear_max ;
	hear_capacity_from_opponent = std->hear_max ;
	pos.x = -(unum * 3 * tm->side) ;
	pos.y = -PITCH_WIDTH/2.0 - 3.0 ;
	angle_neck = 0.0 ;
	angle_neck_committed = 0.0 ;
    angle_body = 0.0 ;

	count_kick = count_dash = count_turn = count_say = count_turn_neck = 0 ;

}

void Player::init(Value version_tmp, int goalie_flag)
{
	version = version_tmp ;
    goalie = goalie_flag ;
    alive = STAND ;
    version = version_tmp ;
    if (goalie)
        alive |= GOALIE ;
    goalie_catch_ban = FALSE ;

	sprintf(name,PLAYER_NAME_FORMAT,team->name,unum) ;
	sprintf(name_far,PLAYER_NAME_FAR_FORMAT,team->name) ;
	sprintf(name_toofar,PLAYER_NAME_TOOFAR_FORMAT) ;

	Set(
		MPO_Player,
		pos,
		stadium->psize,
		SideDirection(team->side),
		PVector(0.0,0.0),PVector(0.0,0.0),
		stadium->pdecay,
		stadium->prand) ;
}

void Player::rotate(Angle a)
{
	angle_body = normalize_angle( angle + a );
}

void Player::rotate_to(Angle dir)
{
	angle_body = normalize_angle(dir) ;
}

void Player::turn(Value a)
{
	Value r1 = vel.r() ;
	rotate((1.0 + drand(-randp, randp)) * a/(1.0 + inertia_moment * r1)) ;
}

void Player::turn_neck(Value a)
{
	angle_neck = NormalizeNeckAngle( angle_neck + a );
}

void Player::kick(Value power, Angle dir) 
{
	MPObject* ball = stadium->ball ;
	PVector tmp ;
	Value dir_diff ;
	Value dist_ball ;

	alive |= KICK ;

	if (stadium->mode == PM_BeforeKickOff ||
		stadium->mode == PM_AfterGoal_Left ||
		stadium->mode == PM_AfterGoal_Right  ||
		stadium->mode == PM_OffSide_Left ||
		stadium->mode == PM_OffSide_Right )
		return ;

	if(pos.distance(ball->pos) > stadium->kickable_area) {
		alive |= KICK_FAULT ;
		return ;
	}

	if (stadium->useoffside)
		if (stadium->mark_offside_mark(this))
			return ;

	dir_diff = fabs(this->vangle(*ball)) ;
	tmp = ball->pos - this->pos ;
	dist_ball = tmp.r() - stadium->psize - stadium->bsize ;

	power *= stadium->kprate *
			(1 - 0.25*dir_diff/PI - 0.25*dist_ball/stadium->kmargin) ;

	ball->push(Polar2PVector(power, dir+angle)) ;
	ball->angle = dir+angle ;
	stadium->last_touch = this ;
	stadium->last_kicker = this ;
	stadium->caught_ball = FALSE ;
}

void Player::goalie_catch(Angle dir) 
{
	static RArea p_l(PVector(-PITCH_LENGTH/2+PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	static RArea p_r(PVector(+PITCH_LENGTH/2-PENALTY_AREA_LENGTH/2.0, 0.0),
					 PVector(PENALTY_AREA_LENGTH, PENALTY_AREA_WIDTH)) ;
	static RArea catchable(PVector(stadium->catch_area_l/2.0, 0.0),
						PVector(stadium->catch_area_l, stadium->catch_area_w)) ;
	PVector	tmp ;
	MPObject* ball = stadium->ball ;

	alive |= CATCH ;

	if (!this->goalie ||
		this->goalie_catch_ban ||
		stadium->mode == PM_BeforeKickOff ||
		stadium->mode == PM_AfterGoal_Left ||
		stadium->mode == PM_AfterGoal_Right ||
		stadium->mode == PM_OffSide_Left ||
		stadium->mode == PM_OffSide_Right )
		return ;

	switch (this->team->side) {
	  case LEFT:
		if (!(p_l.inArea(this->pos))) {
			alive |= CATCH_FAULT ;
			return ;
		}
		break ;
 	  case RIGHT:
		if (!(p_r.inArea(this->pos))) {
			alive |= CATCH_FAULT ;
			return ;
		}
		break ;
	}

	this->goalie_catch_ban = stadium->cban_cycle ;
	tmp = ball->pos - this->pos ;
	tmp.rotate(-(this->angle + dir)) ;
	if (!(catchable.inArea(tmp)) || (drand(1,0) >= goalie_catch_probability)) {
		alive |= CATCH_FAULT ;
		return ;
	}
	relative_ball_ang = ball->pos - this->pos;
	relative_ball_ang /= relative_ball_ang.r();
	stadium->goalie_catch(this) ;
}

Team::Team(Stadium *std, const Side s)
{
	stadium = std ;
	enable = FALSE ;
	name = NULLCHAR ;
	side = s ;
	Max = MAX_PLAYER ; 
	point = 0 ;
	n = 0 ;
	goalie = FALSE ;
	player = new Player*[Max] ;

//	OnlineCoach* olcoach ;
//	olcoach = new OnlineCoach ;
}

Player *Team::newPlayer(Value version, int goalie_flag) 
{
	if (n >= Max) {
		if (stadium->verbose)
			cerr << "Warning:Too many players." << endl ;
		return (Player*)NULL ;
	}

	if (goalie && goalie_flag) {
		if (stadium->verbose)
			cerr << "Warning:Too many goalies." << endl ;
		return (Player*)NULL ;
	}

	if (goalie_flag)
		goalie = TRUE ;

	Player *p = player[n] ;
	n++ ;

	p->init(version, goalie_flag) ;

	return p ;
}

int OnlineCoach::scanField()
{
	PObjectTable *tbl ;
	MPObjectTable *mtbl ;
	int i,j,c ;
	char tmp[16] ;
	PVector etmp, vtmp, pos ;
	float version = 5.0 ;
	Angle ang ;

	if(!vi) 
		vi = new VisualInfo[MaxObject];
		for(int z=0;z<MaxObject;z++) vi[z].set(version);
	c = 0 ;

	pos.x = PITCH_WIDTH/2 ; pos.y = 0 ;

	tbl = &(stadium->votable) ;
	mtbl = &(stadium->motable) ;

	for(i = 0 ; i < tbl->n ; i++ ) {
		if((*tbl->object[i]).enable == FALSE)
			continue ;
		ang = pos.vangle((*tbl->object[i]).pos,angle) ;

		vi[c].highquality = TRUE ;
		vi[c].object = tbl->object[i] ;
		vi[c].name = tbl->object[i]->name ;
		vi[c].distance = pos.distance(tbl->object[i]->pos) ;
		vi[c].direction = ang ;

		if (vi[c].distance == Zero)
			vi[c].dist_change = vi[c].dir_change = Zero ;
		else {
			etmp = tbl->object[i]->pos - pos ;
			sscanf(tbl->object[i]->name, "(%[^ )]", tmp);
			if (!strcmp(tmp, "ball")) {
				vi[c].qstep = stadium->dist_qstep ;
				for(j = 0 ; j < mtbl->n ; j++ ) {
					if (tbl->object[i]->id == mtbl->object[j]->id) {
						vtmp = mtbl->object[j]->vel ;
						break ;
					}
				}
			}
			else if (!strcmp(tmp, "player")) {
				vi[c].qstep = stadium->dist_qstep ;
				for(j = 0 ; j < mtbl->n ; j++ ) {
					if (tbl->object[i]->id == mtbl->object[j]->id) {
						vtmp = mtbl->object[j]->vel ;
						break ;
					}
				}
				vi[c].body_dir = 
					normalize_angle(tbl->object[i]->angle - angle) ;
				vi[c].face_dir = 
					normalize_angle((((Player *)tbl->object[i])->angle_neck_committed
						+ ((Player *)tbl->object[i])->angle) - angle) ;
			}
			else {
				vi[c].qstep = stadium->land_qstep ;
//				vtmp.x = -this->vel.x ;
//				vtmp.y = -this->vel.y ;
				continue ;
			}
			etmp /= vi[c].distance ;
			vi[c].dist_change = vtmp.x * etmp.x + vtmp.y * etmp.y ;
			vi[c].dir_change = RAD2DEG * (vtmp.y * etmp.x
							- vtmp.x * etmp.y) / vi[c].distance ;

			c++ ;
		}
	}

	 // line information
	ang = Rad2IDeg(normalize_angle(angle)) ;
	PObject *dir[2], *dir2[2] ;
	Value edge[2], r[4] ;
	int out_of_field = TRUE ;
	int even = FALSE ;

#ifndef WRONG_LINE_VIEW		/* line view type is normal */
	if (pos.y < -PITCH_WIDTH/2) {	/* above line_t */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
				r[3] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_b) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_t) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > 0 && ang < 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
			}
			else if (ang > 90 && ang < 180) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
			}
			else if (ang == 90) {
				r[0] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_t) ;
				if (ang < 90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_t) ;
				dir2[i] = &(stadium->field.line_b) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;
		}
	}
	else if (pos.y > PITCH_WIDTH/2) {	/* under line_b */
		if (pos.x < -PITCH_LENGTH/2) {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_l) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_r) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else if (pos.x > PITCH_LENGTH/2) {
			if (ang > -180 && ang < -90) {
				r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
				r[3] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				if (r[0] >= r[1] && r[1] < r[3]) {
					dir[i=0] = &(stadium->field.line_r) ;
					dir2[i] = &(stadium->field.line_t) ;
				}
				else if (r[0] < r[1] && r[0] < r[2]) {
					dir[i=1] = &(stadium->field.line_b) ;
					dir2[i] = &(stadium->field.line_l) ;
				}
				else
					return c ;
			}
			else
				return c ;
		}
		else {
			if (ang > -90 && ang < 0) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_r
			}
			else if (ang > -180 && ang < -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ;// to line_l
			}
			else if (ang == -90) {
				r[0] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ; // to line_t
				r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_b
				r[2] = PITCH_LENGTH * 1.5 ;
			}
			else
				return c ;
			
			if (r[1] < r[2] && r[2] < r[0]) {
				dir[i=1] = &(stadium->field.line_b) ;
				if (ang > -90)
					dir2[i] = &(stadium->field.line_r) ;
				else
					dir2[i] = &(stadium->field.line_l) ;
			}
			else if (r[1] < r[2] && r[0] < r[2]) {
				dir[i=1] = &(stadium->field.line_b) ;
				dir2[i] = &(stadium->field.line_t) ;
				r[2] = r[0] ;
				even = TRUE ;
			}
			else
				return c ;

		}
	}
	else if (pos.x < -PITCH_LENGTH/2) {	/* left of line_l */
		if (ang > -90 && ang < 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_t
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
		}
		else if (ang > 0 && ang < 90) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;   // to line_b
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
		}
		else if (ang == 0) {
			r[0] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_l) ;
			if (ang < 0)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_l) ;
			dir2[i] = &(stadium->field.line_r) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else if (pos.x > PITCH_LENGTH/2) {	/* right of line_r */
		if (ang > -180 && ang < -90) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
			r[1] = Abs((-PITCH_WIDTH/2 - pos.y) / sin(angle)) ;  // to line_t
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
		}
		else if (ang > 90 && ang < 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
			r[1] = Abs((PITCH_WIDTH/2 - pos.y) / sin(angle)) ;   // to line_b
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
		}
		else if (ang == 180) {
			r[0] = Abs((PITCH_LENGTH/2 - pos.x) / cos(angle)) ;  // to line_r
			r[1] = PITCH_LENGTH * 1.5 ;
			r[2] = Abs((-PITCH_LENGTH/2 - pos.x) / cos(angle)) ; // to line_l
		}
		else
			return c ;
			
		if (r[0] < r[1] && r[1] < r[2]) {
			dir[i=0] = &(stadium->field.line_r) ;
			if (ang < -90)
				dir2[i] = &(stadium->field.line_t) ;
			else
				dir2[i] = &(stadium->field.line_b) ;
			r[3] = r[1] ;
		}
		else if (r[0] < r[1] && r[2] < r[1]) {
			dir[i=0] = &(stadium->field.line_r) ;
			dir2[i] = &(stadium->field.line_l) ;
			r[3] = r[2] ;
			even = TRUE ;
		}
		else
			return c ;
	}
	else {		/* In the field */
		out_of_field = FALSE ;
		if (ang > -90 && ang <= 90) {
			dir[0] = &(stadium->field.line_r) ;
			edge[0] = PITCH_LENGTH/2 ;
		}
		else {
			dir[0] = &(stadium->field.line_l) ;
			edge[0] = -PITCH_LENGTH/2 ;
		}
		if (ang > -180 && ang <= 0) {
			dir[1] = &(stadium->field.line_t) ;
			edge[1] = -PITCH_WIDTH/2 ;
		}
		else {
			dir[1] = &(stadium->field.line_b) ;
			edge[1] = PITCH_WIDTH/2 ;
		}

		r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle)) ;
		r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle)) ;

		i = (r[0] <= r[1]) ? 0 : 1 ;
	}
#else		/* defined WRONG_LINE_VIEW */
	out_of_field = FALSE ;
	if (ang > -90 && ang <= 90) {
		dir[0] = &(stadium->field.line_r) ;
		edge[0] = PITCH_LENGTH/2 ;
	}
	else {
		dir[0] = &(stadium->field.line_l) ;
		edge[0] = -PITCH_LENGTH/2 ;
	}
	if (ang > -180 && ang <= 0) {
		dir[1] = &(stadium->field.line_t) ;
		edge[1] = -PITCH_WIDTH/2 ;
	}
	else {
		dir[1] = &(stadium->field.line_b) ;
		edge[1] = PITCH_WIDTH/2 ;
	}

	r[0] = (ang == 90 || ang == -90) ?
			PITCH_LENGTH * 1.5 : Abs((edge[0] - pos.x) / cos(angle)) ;
	r[1] = (ang == 0 || ang == 180) ?
			PITCH_LENGTH * 1.5 : Abs((edge[1] - pos.y) / sin(angle)) ;
	i = (r[0] <= r[1]) ? 0 : 1 ;
#endif

	if (i == 0)
		ang = -angle + ((angle > 0) ? PI/2 : -PI/2) ;
	else
		ang = -angle ;
	if (ang > PI/2) ang -= PI ;
	if (ang < -PI/2) ang += PI ;

	vi[c].highquality = TRUE ;
	vi[c].object = dir[i] ;
	vi[c].name = dir[i]->name ;
	vi[c].distance = r[i] ;
	vi[c].direction = ang ;
	vi[c].qstep = stadium->land_qstep ;
	c++ ;

	return c ;

}



